#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_PCD8544.h>
#include <si5351.h>
#include "Wire.h"
#include <JTEncode.h>
#include <TimeLib.h>
#include <NTPtimeESP.h>
#include <ESP8266WiFi.h>
#include <WiFiClient.h>

Adafruit_PCD8544 display = Adafruit_PCD8544(D7, D6, D5, D4, D3);


#define TONE_SPACING            146           // ~1.46 Hz
#define WSPR_DELAY              683          // Delay value for WSPR
#define WSPR_CTC                10672         // CTC value for WSPR
#define SYMBOL_COUNT            WSPR_SYMBOL_COUNT
#define CORRECTION              163800             // Change this for your ref osc -12000 

#define TX_LED_PIN              2  //integrated onboard led, marked as D4 on Wemos D1 mini lite
#define SYNC_LED_PIN            16 //marked as D on Wemos D1 mini lite

//****************************************************
//* SYNCRONIZE SYSTEM TIME with NTP SERVERS
//* need to be modified, obsolete library in use...
//****************************************************
#define SEND_INTV     10
#define RECV_TIMEOUT  10

//****************************************************
//list of frequencies for jack 1 (clock0 of si5351) output
#define MAXCH      8 //Change this according to the number of bands inserted below
unsigned long freq0[] = {28126100UL, 24926100UL, 21096100UL, 18106100UL, 14097100UL, 10140200UL, 7040100UL, 3570100UL}; //CHANGE THIS: is the freq of multiband output on jack 1 (clock0)


#define SI5351_REF     27000000UL  //change this to the frequency of the crystal on your si5351’s PCB, usually 25 or 27 MHz

char call[7] = "RA4FHE";                        // Change this
char loc[5] = "LO23";                           // Change this
uint8_t dbm = 10;
uint8_t tx_buffer[SYMBOL_COUNT];

const char* ssid = "Infotex";       //SSID of your Wifi network: Change this
const char* password = "12345678";      //Wi-Fi Password:            Change this

//**** How the station is named in your NET
const char* WiFi_hostname = "WSPR_HF";

//**** Sync the soft clock every 12 hours
#define NTPSYNC_DELAY  6

//**** NTP Server to use
const char* NTP_Server = "ntp1.stratum2.ru";

//**** Your time zone UTC related (floating point number)
#define TIME_ZONE 3.0f

Si5351 si5351;
//some system variable. Anything to touch from here
// Global variables
JTEncode jtencode;
int ch = 0; // automatic frequency switch index
bool warmup = 0;

NTPtime NTPch(NTP_Server);
strDateTime dateTime;

// --------------------------------------
// epochUnixNTP set the UNIX time
// number of seconds sice Jan 1 1970
// --------------------------------------

time_t epochUnixNTP()
{
  Serial.println(">>>>>>>> Time Sync function called <<<<<<<<<");

  //**** BIG ISSUE: in case of poor connection, we risk to remain in this loop forever
  NTPch.setSendInterval(SEND_INTV);
  NTPch.setRecvTimeout(RECV_TIMEOUT);
  do
  {
    dateTime = NTPch.getNTPtime(TIME_ZONE, 1);
    delay(1);
  }
  while (!dateTime.valid);
  NTPch.printDateTime(dateTime);
  setTime(dateTime.hour, dateTime.minute, dateTime.second, dateTime.day, dateTime.month, dateTime.year);
  Serial.println(now());
  return 0;
}

void printTx(bool tx)
{
  display.setTextSize(1);
  display.fillRect(0, 40, 70, 8, WHITE);
  display.setTextColor(BLACK);
  display.setCursor(0, 40);
  if (tx) {
    display.print("TX ON");
  } else {
    display.print("TX OFF");
  }
  display.display();
}

void printFreq()
{
  display.setTextSize(1);
  display.fillRect(0, 30, 75, 8, WHITE);
  display.setTextColor(BLACK);
  display.setCursor(0, 30);
  display.print(freq0[ch]);
  display.println(" Hz");
  display.display();
}

void printTime(bool printSync)
{
  display.setTextSize(1);
  display.fillRect(0, 0, 84, 10, WHITE);
  display.setTextColor(BLACK);
  display.setCursor(0, 0);
  if (printSync) {
    display.print("Sync time...");
  } else {
    display.printf("%02d:%02d:%02d", hour(), minute(), second());
  }
  display.display();
}

// Loop through the string, transmitting one character at a time.
void encode()
{
  uint8_t i;

  jtencode.wspr_encode(call, loc, dbm, tx_buffer);

  si5351.set_clock_pwr(SI5351_CLK0, 1);
  
  digitalWrite(TX_LED_PIN, LOW);
  Serial.println("TX ON");
  printTx(true);
  printFreq();

  // Now do the rest of the message
  uint32_t period = millis() - WSPR_DELAY;
  uint8_t sec = second();
  i = 0;
  while (i < SYMBOL_COUNT) {
    uint32_t curMill = millis();
    if (curMill - period >= WSPR_DELAY) {
      period = curMill;
      si5351.set_freq((freq0[ch] * 100) + (tx_buffer[i] * TONE_SPACING), SI5351_CLK0);
      ++i;
    }
    if (sec != second()) {
      sec = second();
      printTime(false);
    }
  }

  // Turn off the output
  si5351.set_clock_pwr(SI5351_CLK0, 0);

  digitalWrite(TX_LED_PIN, HIGH);
  Serial.println("TX OFF");
  printTx(false);

  //change band on clock0 for next tx
  ch++;
  if (ch == MAXCH) ch = 0; //reset band index to start again from first band the in array
  printFreq();
}

void printWiFi()
{
  display.setTextSize(1);
  display.fillRect(0, 10, 84, 20, WHITE);
  display.setTextColor(BLACK);
  display.setCursor(0, 10);
  if (WiFi.status() == WL_CONNECTED) {
    display.print("WI-FI: OK");
  } else {
    display.print("WI-FI: ERR");
  }
  display.setTextColor(BLACK);
  display.setCursor(0, 20);
  display.println(WiFi.localIP());
  display.display();
}

void ssidConnect()
{
  Serial.println(ssid);
  Serial.println(password);
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(700);
    Serial.print(".");
  }

  Serial.println();
  Serial.print(F("Connected to "));
  Serial.println(ssid);
  display.clearDisplay();
  printWiFi();
  Serial.print(F("IP address: "));
  Serial.println(WiFi.localIP());
}

void setup()
{
  Serial.begin(74880); while (!Serial);
  Serial.println("COM setup successful");
  delay(10);
  display.begin();
  display.setContrast(58);
  display.clearDisplay();

  display.setTextSize(1);
  display.setTextColor(BLACK);
  display.setCursor(15, 20);
  display.print("Loading...");
  display.display();
  delay(1000);

  display.clearDisplay();
  display.setCursor(15, 14);
  display.print("Connect to");
  display.setCursor(21, 24);
  display.print("WiFi...");
  display.display();

  WiFi.mode(WIFI_STA);
  // connect to WiFi network
  ssidConnect();

  // Use the LED as a keying indicator.
  pinMode(TX_LED_PIN, OUTPUT);
  pinMode(SYNC_LED_PIN, OUTPUT);
  digitalWrite(TX_LED_PIN, HIGH);
  digitalWrite(SYNC_LED_PIN, HIGH);

  printTime(true);

  // Set time sync provider
  setSyncProvider(epochUnixNTP);  //set function to call when sync required
  setSyncInterval(3600 * 24); // Sync time every 24 hour

  // Initialize the Si5351
  // Change the 2nd parameter in init if using a ref osc other
  // than 25 MHz
  Serial.println("start radio module setup");
  si5351.init(SI5351_CRYSTAL_LOAD_8PF, SI5351_REF, CORRECTION);
  Serial.println("Module intializated");

 // Set CLK0 out  put
  si5351.set_freq(freq0[ch] * 100, SI5351_CLK0);
  si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_8MA); // Set for max power
  si5351.set_clock_pwr(SI5351_CLK0, 0); // Disable the clock initially
  Serial.println("Radio Module setup successful");
  Serial.println("Entering loop...");
}

void loop()
{
  printTime(false);
  printWiFi();

  if (minute() % 2 == 0 && second() == 0)
  {
    encode();
  }
  delay(50);
}

